#####################################################
#  HOA for COX Regression  - version 0.50           #
#####################################################
#
#  - version without stratification
#  - works when MLE has been computed by a call to coxph()
#  - uses Efron method for ties 
#  - requires latest version of survival package
#

####################################################
# fitting functions: partial lik score and hessian #
####################################################
# partial log-likelihod (sign changed)
parLik.c<-function(beta,fo,Zo,oo)
{   
	 out<-.C("par_lik",
     as.double(beta),
     as.integer(length(beta)),
     as.double(fo),
     as.integer(length(fo)),
     as.double(Zo),
     as.double(oo),
     l=as.double(0.0),PACKAGE="hoacoxph")$l
     -out
}


#partial likelihood score function (sign changed)
scoreLik.c<-function(beta,fo,Zo,oo)
{ 
 out<- .C("score_lik",
     as.double(beta), 
     as.integer(length(beta)),
     as.double(fo),
     as.integer(length(fo)),
     as.double(Zo),
     as.double(oo),
     s=as.double(rep(0.0,length(beta))),PACKAGE="hoacoxph")$s
    -out
     }
 
#maximizes the partial likelihood
maxLik<- function(fo,Zo,oo=rep(0,nrow(Zo)),beta.init=rep(0,ncol(Zo)),...)
{
   n<-length(fo)
   obj<- try(coxph.fit(Zo,cbind(1:n,fo),offset=oo,strata=NULL,control=coxph.control(...),
                  method="efron",rownames=NULL,init=beta.init))
   if(is.character(obj)) 
     {obj <- optim(beta.init,parLik.c,fo=fo,Zo=Zo,oo=oo)
      out <- list(par=obj$par, value=obj$value)
     }         
    else  out<- list(par=obj$coef, value=parLik.c(obj$coef,fo,Zo,oo) )
    return(out)
  }
  
     
  
# information matrix
infoLik.c<-function(beta,fo,Zo,oo)
{ 
	out<-.C("info_lik",
     as.double(beta), 
     as.integer(length(beta)),
     as.double(fo),
     as.integer(length(fo)),
     as.double(Zo),
     as.double(oo),
     H=as.double(rep(0.0,length(beta)^2)),PACKAGE="hoacoxph")$H 
     out<-matrix(out,nrow=length(beta))
     return(out)
}




####################################################
#  HIGHER ORDER METHODS                            #
####################################################


# performs the loop in ASampling
adaptCens.c<-function(f,RS)
{ out<- .C("adapt_cens",
      as.double(f),
      as.double(RS),
      as.integer(length(f)),
     ind=as.double(f),PACKAGE="hoacoxph")$ind
     out
}

# resampling based on reference censoring model
Asampling<-function(fo,Zo,beta)
{
  n <- length(fo)	
  RS <- 1:n ### they identifies elements of [Zo,fo]
  out <- rep(NA,n)
  lambda <- exp(Zo %*% beta)
  y.sim <- -log(1-runif(n)) / lambda
  ord.sim <- order(y.sim)
  RS.sim <- RS[ord.sim]
  out <- adaptCens.c(fo,RS.sim)
  return(out)
}

  

# computes P(r(psi)<=r_obs) by bootstrap based on a reference censoring model
coxph.sim<-function(coxph.obj, psi, ind.psi, R=10000, trace=TRUE)
{
  mle<- update(coxph.obj, method="efron")	
  r.vals<- rep(0,R)
  y<- mle$y[,1]
  f<- mle$y[,2]
  ord <- order(y)
  fo <- f[ord]
  p<- length(mle$coef)
  n<- length(y)
  Z<- model.matrix(mle)
  it.trouble<- 0
  if(p==1)
     { Z<- as.matrix(Z)
       Zo<- as.matrix(Z[ord])
       l1<-  parLik.c(mle$coef,fo,Zo,rep(0,n))
       l0<-  parLik.c(psi,fo,Zo,rep(0,n))
       r.obs<-  sqrt(2*(-l1+l0))*sign(mle$coef-psi)
        for (i in 1:R)
        {  if((trace) & (i%%(R/10)==0)) cat("sim. numb.",i,"\n")
           ord.sim<- Asampling(fo,Zo,psi)
           Z.sim <- as.matrix(Zo[ord.sim,])
           obj.hyp<-  parLik.c(psi,fo,Z.sim,rep(0,n))
           obj.hat<- maxLik(fo,Z.sim,beta.init=psi)
           r.vals[i]<- sqrt(2*(-obj.hat$val+obj.hyp))*sign(obj.hat$par-psi)
         }
     }
 if(p>1)
   { Zo <- Z[ord,]
   	 ind.nu<- setdiff(1:p, ind.psi)
     mle.hyp<- coxph(Surv(y,f)~Z[,ind.nu]+offset(Z[,ind.psi]*psi), method="efron")
     r.obs<-  sqrt(2*(mle$loglik[2]-mle.hyp$loglik[2]))*sign(mle$coef[ind.psi]-psi)
     beta.hyp<- rep(0,p)
     beta.hyp[ind.psi] <- psi
     beta.hyp[ind.nu] <- mle.hyp$coef
     for (i in 1:R)
        {  if((trace) & (i%%(R/10)==0))  cat("sim. numb.=",i,"\n")
           ord.sim<- Asampling(fo,Zo,beta.hyp)
           Z.sim <- Zo[ord.sim,]
           obj.hyp <- maxLik(fo,as.matrix(Z.sim[,-ind.psi]),psi*Z.sim[,ind.psi])
           beta.init <- rep(psi,p)
           beta.init[-ind.psi] <- obj.hyp$par
           obj.hat<- maxLik(fo,Z.sim,beta.init=beta.init)
           r.vals[i]<- sqrt(2*(-obj.hat$val+obj.hyp$val))*sign(obj.hat$par[ind.psi]-psi)
           if(is.na(r.vals[i])) #### r=NA taken as monotone likelihood case  
             { r.vals[i]<-0
                it.trouble<- it.trouble + 1
             }
          }
    }
  if(trace) { hist(r.vals)
              abline(v=r.obs,col=2)}
  out<- list(r.obs=r.obs, r.sim=r.vals, it.trouble=it.trouble)
  return(out)
  }
  
  
  


#### high-order asymptotics
coxph.hoa<-function(coxph.obj, psi, ind.psi, R=3000, fixed.risk.set=FALSE,tol=10^-5)
 {
 	
 	
  likscoreLik.c<-function(beta0, beta1, fo,Zo)
  { .C("likscore_lik",
     as.double(beta0),
     as.double(beta1),
     as.integer(length(beta0)),
     as.double(fo),
     as.integer(length(fo)),
     as.double(Zo),
     s=as.double(rep(0.0,2*(length(beta0)+1))))$s
   }


 	
  getSimMat<- function(beta.hat,beta.hyp,Zo,fo,R)
  { p<-length(beta.hat)
    U.hat <- U.til<- matrix(0, R, p)
    ldrop <- rep(0, R)
    for (i in 1:R)
      { 
         ord.sim<- Asampling(fo,Zo,beta.hat)
         Z.sim <- if(p>1) Zo[ord.sim,]
         else as.matrix(Zo[ord.sim,])
         ls.obj<-  likscoreLik.c(beta.hat,beta.hyp,fo,Z.sim)
         U.hat[i,] <- ls.obj[1:p]
         U.til[i,] <- ls.obj[(p+2):(2*(p+1)-1)]
         ldrop[i] <- ls.obj[p+1] - ls.obj[p+1+p+1]
        }
     return(list(U.hat=U.hat, U.til=U.til, ldrop=ldrop))
  }

  mle<- update(coxph.obj, method="efron")
  y<- mle$y[,1]
  f<- mle$y[,2]
  ord <- order(y)
  fo <- f[ord]
  p<- length(mle$coef)
  n<- length(y)
  Z<- model.matrix(mle)
  if(p==1)
     {
      Z<- as.matrix(Z)
      Zo<- as.matrix(Z[ord])
      l1<-  parLik.c(mle$coef,fo,Zo,rep(0,n))
      l0<-  parLik.c(psi,fo,Zo,rep(0,n))
      r.obs<-  sqrt(2*(-l1+l0))*sign(mle$coef-psi)
      info.f<- infoLik.c(mle$coef,fo,Zo,rep(0,n))
      se <- 1/info.f^0.5
      INF<-0
      NP<- NA
      w<- (mle$coef-psi)/se 	
      if ((!fixed.risk.set) & (abs(r.obs)>tol)) 
         {
     	  obj<- getSimMat(mle$coef,psi,Zo,fo,R)
          U.hat<- obj$U.hat
          ldrop<- obj$ldrop
          i.hat <-  cov(U.hat)
          q.psi <- cov(ldrop, U.hat)
          INF <- 1/r.obs*log(q.psi/i.hat*info.f^0.5/r.obs)
     	   }
     if ((fixed.risk.set) & (abs(r.obs)>tol)) INF <- 1/r.obs*log(w/r.obs)	     
     } 	   
   if(p>1)
   { Zo <- Z[ord,]
   	 ind.nu<- setdiff(1:p, ind.psi)
     mle.hyp<- coxph(Surv(y,f)~Z[,ind.nu]+offset(Z[,ind.psi]*psi), method="efron")
     r.obs<-  sqrt(2*(mle$loglik[2]-mle.hyp$loglik[2]))*sign(mle$coef[ind.psi]-psi)
     if(is.na(r.obs)) r.obs<-0
     info.f<- infoLik.c(mle$coef,fo,Zo,rep(0,n))
     se <- solve(info.f)[ind.psi, ind.psi]^0.5
     NP<-INF<-0
     beta.hyp<- beta.hat<- mle$coef
     beta.hyp[ind.psi] <- psi
     beta.hyp[ind.nu] <- mle.hyp$coef
     w<- (mle$coef[ind.psi]-psi)/se
      if ((!fixed.risk.set) & (abs(r.obs)>tol))  
        {
         obj<- getSimMat(beta.hat,beta.hyp,Zo,fo,R)
         U.hat<- obj$U.hat
         U.til<- obj$U.til
         ldrop<- obj$ldrop
         cov.all <-  cov(cbind(U.hat, U.til))
         S<-   cov.all[(p+1):(2*p), 1:p]
         i.hat <- cov.all[1:p, 1:p]
         info.h<- infoLik.c(beta.hyp,fo,Zo,rep(0,n))[ind.nu,ind.nu]
         q <- cov( cbind(ldrop, U.hat))[1, 2:(p+1)]
         SS2 <- S %*% solve(i.hat) %*% info.f
         SS1 <- q %*% solve(i.hat) %*% info.f
         numU <- SS1[ind.psi] - t(SS2[-ind.psi,ind.psi]) %*% solve(SS2[-ind.psi,-ind.psi]) %*% SS1[-ind.psi]
         jHatInv <- solve(info.f)  
         jProf <- 1 / jHatInv[ind.psi,ind.psi] 
         u <- numU / sqrt(jProf)
         CPsi <- det(as.matrix(SS2[-ind.psi,-ind.psi])) / sqrt(det(as.matrix(info.h)) * det(as.matrix(info.f[-ind.psi,-ind.psi])))  
         NP <- (1/r.obs) * log(CPsi)  
         INF <- (1/r.obs) * log(u/r.obs)   
         }
     if ((fixed.risk.set) & (abs(r.obs)>tol)) 
        {
         info.h<- infoLik.c(beta.hyp,fo,Zo,rep(0,n))[ind.nu,ind.nu]
         INF <- 1/r.obs*log(w/r.obs)
         NP <- 1/r.obs*log( det(as.matrix(info.f[ind.nu,ind.nu]))^0.5 / sqrt(det(as.matrix(info.h))))
         }    
           
     }
     NP.fin<- if(is.na(NP)) 0 else NP
     out<-list(wald=w, r.obs=r.obs, NP=NP, INF=INF,  rs.obs=r.obs+INF+NP.fin)
     return(out)
   	}  




